//3456789_123456789_123456789_123456789_123456789_123456789_123456789_123456789_


//#include <stdafx.h>				// for VC++ precompiled headers.
#include <math.h>					// For sines, cosines
#include <stdio.h>
#include <string.h>					// for prnt() and textOut() routines.
#include <stdlib.h>					// for rand(), srand()
#include <stdarg.h>					// for variable argument lists.
#include <assert.h>					// for non-Windows 
#include <time.h>					// for seeding the random number gen.
#include "v_vecmat.h"				// for vector/matrix classes

#define ASSERT assert				// error trap
#define DIAGNOSTICS TRUE			// helpful ordinary checking.
#define PESKYWARNINGS FALSE			// tediously tight checking.
//#define dprintf TRACE				// diagnostic printf; (from Visual C++2.0)
									//      set to TRACE for VC++ diag. window,
#define dprintf printf				//		set to 'printf' for UNIX/cmd-line.
									

//==============================================================================
//========================                ======================================
//========================   2x1 Vectors  ======================================
//========================                ======================================
//==============================================================================
// All member functions use internal tmp. storage, you can use the'this'
// keyword as a function argument without problems.

#undef VV
#define VV 2					// 2 element vectors, 2x2 element matrix

//======================== Vec2 Member fcns ====================================
//========================------------------====================================

void Vec2::prnt(void) { 
//------------------------------------------------------------------------------
// Unlabeled prettyprint of Vec2 elements.
char buf[81];

    DPF("\n--VEC2:--\t");
	sprintf(buf,"%f\t%f\n",row[0],row[1]);
	DPF(buf);						// (DPF can't do floats (Jerks!)
}

void Vec2::prnt(char *name) {
//------------------------------------------------------------------------------
// Labeled prettyprint of Vec2 elements.
char buf[81];

	sprintf(buf,"\n%s: %f\t%f\n",name,row[0],row[1]);
	DPF(buf);
}

VECTYPE Vec2::norm(Vec2& src) {
//------------------------------------------------------------------------------
//Replace 'self' with unit-length version of 'src' vector; retn length.
int i;
VECTYPE length = (VECTYPE)0.0;
VECTYPE tmp    = (VECTYPE)0.0;

    for(i=0; i< VV; i++) 					// get vector length.
	{
		length += src.row[i] * src.row[i];
	}
    if(length==(VECTYPE)1.0) 
	{	
		*this = src;
		return((VECTYPE)1.0);				// already normalized.
	}

    length = (VECTYPE)sqrt((double)length);

    if(length != (VECTYPE)0.0)				// normalize it.	
	{
		tmp = (VECTYPE)1.0/ length;
		for(i=0; i< VV; i++) 
	 	{
		   row[i] = tmp * src.row[i];
	 	}
		return(length);
		}
    else {
#if(DIAGNOSTICS==TRUE)
		DPF("\n ???Vec2::norm() given 0-length src vector!!!");
#endif
		return((VECTYPE)0.0);
	}
}

void Vec2::scale(Vec2& src1,VECTYPE c) 
//------------------------------------------------------------------------------
// Multiply by scalar: [this]= src1 *c
{
int i;
Vec2 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i]*c;
	}
    *this = out;	
}

void Vec2::add(Vec2& src1, Vec2& src2) 
//------------------------------------------------------------------------------
// Vector addition: [this]= src1 + src2
{
int i;
Vec2 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i] + src2.row[i];
	}
    *this = out;	
}

void Vec2::addIn(Vec2& src) 
//------------------------------------------------------------------------------
// Vector addition: [this]= src + [previous_this]
{
int i;
Vec2 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = this->row[i] + src.row[i];
	}
    *this = out;				// copy the result to 'this'
}

void Vec2::sub(Vec2& src1, Vec2& src2) 
//------------------------------------------------------------------------------
// subtract vectors: [this] = src1 - src2
{
int i;
Vec2 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i] - src2.row[i];
	}
    *this = out;
}

VECTYPE Vec2::dot(Vec2& src1, Vec2& src2) 
//------------------------------------------------------------------------------
// vector dot product: [this] = src1 <dot> src2 	
{
VECTYPE out=(VECTYPE)0.0;
int i;

    for(i=0; i <VV; i++) 
	{
		out += src1.row[i] * src2.row[i];
	}
    return(out);    
}

VECTYPE Vec2::cross(Vec2& src1) 
//------------------------------------------------------------------------------
// vector 2D (planar) cross-product: return value = [this] <cross> src1.
// result is |this|*|src1|*sin(angle), where 'angle' measured from this to src1.
{
    return(row[0] * src1.row[1] - row[1] * src1.row[0]);
}

void Vec2::mpy(Mat2& srcM, Vec2& srcV)
//-----------------------------------------------------------------------------
//  Matrix-Vector multiply: [this]=[srcM][srcV]	
{
int i,j;
Vec2 dest;							// local store for result;

    for(j=0; j<VV; j++) 
	{
		for(i=0; i<VV; i++) 
		{
	    	dest.row[j] += srcM.colrow[i][j] * srcV.row[i];
	    }
	}
    *this = dest;	 				// copy results to calling Vec2

}

BOOL Vec2::isCloserTo(Vec2& srcA, Vec2& srcB)
//-----------------------------------------------------------------------------
// TRUE iff distance to 'srcA' is LESS THAN distance to 'srcB' when measured
// from 'this'.  (FALSE if equal distance).
{
VECTYPE dA,dB;
Vec2 tmp;

	tmp.row[0] = srcA.row[0] - row[0];
	tmp.row[1] = srcA.row[1] - row[1];
	dA = tmp.row[0]*tmp.row[0] + tmp.row[1]*tmp.row[1];
	tmp.row[0] = srcB.row[0] - row[0];
	tmp.row[1] = srcB.row[1] - row[1];
	dB = tmp.row[0]*tmp.row[0] + tmp.row[1]*tmp.row[1];
	if(dA<dB) return(TRUE);
	else return(FALSE);
}

VECTYPE Vec2::CCWAngleTo(Vec2& dest)
//-----------------------------------------------------------------------------
// Find angle measured CCW in radians (0-to-2pi) from 'this' to 'dest'.
{
Vec2 n0,n1;
VECTYPE tcos,tsin,out;

	n0.norm(*this);							// make normalized vectors,
	n1.norm(dest);
	tcos = n0.dot(n0,n1);					// n0 dot n1 = cos(theta)
											// n0 cross n1= sin(theta)
	tsin= n0.row[0]*n1.row[1] - n0.row[1]*n1.row[0]; 
	out = (VECTYPE)atan2((double)tsin,(double)tcos);
											// atan2(x,y) returns arctan(y/x)
											// properly, even if x=0.
	if(out<(VECTYPE)0.0) out = (VECTYPE)(M_PI*2.0)+out; 
	return(out);							// convert negative angles to CCW
}

VECTYPE Vec2::signedAngleTo(Vec2& dest) 
//-----------------------------------------------------------------------------
// Find angle in signed radians (-pi to +pi) from 'this' to 'dest'.
{
Vec2 n0,n1;
VECTYPE tcos,tsin,out;

	n0.norm(*this);							// make normalized vectors,
	n1.norm(dest);
	tcos = n0.dot(n0,n1);					// n0 dot n1 = cos(theta)
											// n0 cross n1= sin(theta)
	tsin= n0.row[0]*n1.row[1] - n0.row[1]*n1.row[0]; 
	out = (VECTYPE)atan2((double)tsin,(double)tcos);
											// atan2(x,y) returns arctan(y/x)
											// properly, even if x=0.
	return(out);
}

/*
BOOL Vec2::isInsideQuad(Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3)
//------------------------------------------------------------------------------
// Return TRUE if our contents (the 'pThis' point) describes an x,y point that 
// is inside a quadrilateral bounded region whose corner points are given in 
// CCW order by p0,p1,p2,p3. 
{
Vec2 bnd;						// boundary vector; one side of the quad.
Vec2 tst;						// test vector; from start of bnd to 'pThis'

	bnd.sub(p1,p0);				// first side: (p1 - p0)
	tst.sub(*this,p0);			// vector from p0 to our test point;
	if(bnd.cross(tst) >=(VECTYPE)0.0)	// is test point inside the quad?
	{							// YES! test 2nd side: (p2 - p1)
		bnd.sub(p2,p1);
		tst.sub(*this,p1);		// vector from p1 to our test point;
		if(bnd.cross(tst) >=(VECTYPE)0.0) // is test point inside the quad?
		{						// YES! test 3rd side: (p3 - p2)
			bnd.sub(p3,p2);
			tst.sub(*this,p2);	// vector from p2 to our test point;
			if(bnd.cross(tst) >= (VECTYPE)0.0)	// is test point inside quad?
			{					// YES! test 4th side; (p0 - p3)
				bnd.sub(p0,p3);
				tst.sub(*this,p3);	// vector from p3 to our test point;
				if(bnd.cross(tst) >= (VECTYPE)0.0)	// is test point inside?
				{
					return(TRUE);	// YES! we are inside all 4 sides! DONE!
				}
			}
		}
	}
	return(FALSE);				// nope! we're outside 1 or more boundaries.
}
*/
BOOL Vec2::isInsideQuad(Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3)
//------------------------------------------------------------------------------
// Return TRUE if our contents (the 'pThis' point) describes an x,y point that 
// is inside a quadrilateral bounded region whose corner points are given in 
// CCW order by p0,p1,p2,p3. 
// Method: Draw a ray from point pThis upwards, in the +y direction. How many ]
// sides of the quadrilateral does that ray cross?
//  0,2, or 4 Sides => pThis is outside the quadrilateral.
//   1 or 3 Sides   => pThis is inside the quadrilateral. 
//		(3 sides? 4 sides? we have a convex shape--a chevron or bow-tie.)
{
Vec2 *pA, *pB;					// ptrs to start & end of a side of the quad.
Vec2 pt;						// parameterized position along a side.
VECTYPE t0,denom,yval;			// parameter value where pt.x = pThis->x
int i,cnt;						// side-crossing counter.

	cnt = 0;
	if(this->row[0] > 1.0 && this->row[1] > 0.5)
	{
		cnt = 0;
	}
	for(i=0; i<4; i++)			// for each side of the quad,
	{
		switch(i)
		{
		case 0:	pA = &p0; pB = &p1; break;	// point to endpoints of 1st side;
		case 1:	pA = &p1; pB = &p2;	break;	// 2nd side,
		case 2: pA = &p2; pB = &p3;	break;	// 3rd side,
		case 3: pA = &p3; pB = &p0;	break;	// 4th side.
		default: printf("?!?! out-of-bounds for loop?!?\n"); break;
		}
		////  Ray Test:
		// One-by-one, parameterize each side of the quad by 't': 
		//		p(t) = pA + t(pB-pA),  or pt.x = pA->x + t(pB->x -pA->x)
		//								  pt.y = pA->y + t(pB->y -pA->y)
		// Next, solve for the 't' value t0 where pt.x = pThis->x:
		//		pThis->x = pA->x + t0(pB->x - pA->x),  
		//						so  t0 = (pThis->x - pA->x) / (pB->x - pA->x).
		denom = pB->row[0] - pA->row[0];	
		if(denom*denom < 1E-20)		// If denominator is 10^-10 or smaller,
			continue;				// then this side is parallel to ray, and
									// it doesn't intersect. Go on to next side.
		t0 = (this->row[0] - pA->row[0]) / denom;	// otherwise, find t0.
		if(t0<0.0 || t0 >1.0)		// If both pA and pB are on the same side;
			continue;				// of the ray, it doesn't intersect. Go on.
		//	But if 0.0 <= t0 <=1.0, we need to look further: 
		// find the pt.y value when t=t0: 	
		//		pt.y = pA->y + (pB->y - pA->y)(t0).
		yval = pA->row[1] + (pB->row[1] - pA->row[1])*t0;
		//	if pt.y < pThis->y, then the side crosses the ray BELOW the point 
		//  where the upwards ray  begins, so NO intersection with this side.
		//	else if pt.y >= pThis->y, then we DO have an intersection.
		if(yval >= this->row[1]) cnt++;
	}
	return(cnt%2);					// return TRUE if cnt==1,3,5, ...
									// return FALSE if cnt==0,2,4, ...
}

#define JTQUAD_DEGEN		0
#define JTQUAD_BOWTIE		1
#define JTQUAD_CHEVRON_02	2
#define JTQUAD_CHEVRON_13	3
#define JTQUAD_TRI_012		4
#define JTQUAD_TRI_123		5
#define JTQUAD_TRI_230		6
#define JTQUAD_TRI_301		7
#define JTQUAD_CONVEX		8

int Vec2::classifyQuad(Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3)
//------------------------------------------------------------------------------
// Return a 'classification' code for the quadrilateral bounded region whose 
// corner points are listed in CCW order by p0,p1,p2,p3.
//  code>0 if front-facing, <0 if back-facing, and:
//	|code|	= 8 if the quadrilateral is convex,
//			= 7 if it is a triangle where p3,p0,p1 are collinear;
//			= 6 if it is a triangle where p2,p3,p0 are collinear;
//			= 5 if it is a triangle where p1,p2,p3 are collinear;
//			= 4 if it is a triangle where p0,p1,p2 are collinear;
//			= 3 if it is a 'chevron' where p1,p3 have acute angles,
//			= 2 if it is a 'chevron' where p0,p2 have acute angles,
//			= 1 if it is a 'bowtie' where two opposite edges cross,
//			= 0 if it is degenerate (all 4 points collinear).
//
// Method: 
// Define two diagonal lines:
//		L02 from point p0 along direction vector v02 = p2 - p0;
//		L13 from point p1 along direction vector v13 = p3 - p1;
// Make these tests:
//		if (p0, p1, p2, p3 are collinear, 
//				then STOP: quad is degenerate; zero area.
//		if (p1 is collinear with L02) then we have a triangle;
//						(ignore p1; if it is not between p0 and p2 it
//						causes a zero-area 'hair' extending from a vertex)
//		{
//			if((p2-p0) x (p3-p0) > 0) STOP: quad is front-facing triangle
//			else STOP: quad is a back-facing triangle.
//		}
//		if(p3 is collinear with L02) then we have a triangle:
//						(ignore p3; if it is not between p0 and p2 it
//						causes a zero-area 'hair' extending from a vertex)
//		{
//			if((p2-p0) x (p1-p0) > 0) STOP: quad is a back-facing triangle.
//			else STOP: quad is a front-facing triangle.
//		}
// At this point, we know that neither p0 nor p3 are collinear with L02.
//		if(p0 is collinear with L13) then we have a triangle:
//						(ignore p0; if it is not between p1 and p3 it
//						causes a zero-area 'hair' extending from a vertex)
//		{
//			if((p3-p1) x (p2-p1) > 0) STOP: quad is a back-facing triangle.
//			else STOP: quad is a front-facing triangle.
//		}
//		if(p2 is collinear with L13) then we have a triangle:
//						(ignore p2; if it is not between p1 and p3 it
//						causes a zero-area 'hair' extending from a vertex)
//		{
//			if((p3-p1) x (p0-p1) > 0) STOP: quad is a front-facing triangle.
//			else STOP: quad is a back-facing triangle.
//		}
//	At this point, we know that we don't have a triangle of any sort; we have
//	either a quad, a chevron, or a bow-tie.  To find out which, do these tests
//		if (p1 and p3 are on opposite sides of line L02), c02 = TRUE;
//	    else they are on the SAME side, so set c02 = FALSE.
//		if (p0 and p2 are on opposite sides of line L13) c13 = TRUE;
//		else they are on the SAME side, so set c13 = FALSE.
//	Use this chart:
//		 c02     c13
//		------+------
//		TRUE    TRUE	quad is convex. 
//						if ((p2-p1) x (p3-p1) > 0) STOP: front-facing quad.
//						 else STOP: back-facing quad.
//							
//		TRUE    FALSE	quad is a chevron with acute angles at p1,p3.
//						if((p2-p1) x (p0-p1) >0) STOP: front-facing chevron
//						else STOP: back-facing chevron
//
//		FALSE   TRUE	quad is a chevron with acute angles at p0,p2.
//						if((p1-p0) x (p3-p0) >0) STOP: front-facing chevron
//						else STOP: back-facing chevron.
// 
//		FALSE   FALSE	quad is a 'bow-tie' with crossing edges.
{
Vec2 L02;						// Diagonal line vector: from p0 to p2
Vec2 v01,v03;					// quad-side vectors from p0 (to p1, to p3)
VECTYPE val01,val03;			// cross-product results for L02.
Vec2 L13;						// Diagonal line vector: from p1 to p3
Vec2 v10,v12;					// quad-side vectors from p1 (to p0, to p2
VECTYPE val10,val12;			// cross-product results for L13.
int c02, c13;					// straddle/same side flags.

	L02.sub(p2,p0);				// make the L02 vector from p0 to p2.
	v01.sub(p1,p0);				// make quad-side vectors from p0
	v03.sub(p3,p0);
	val01 = L02.cross(v01);		// test p1: left, right, or on top of line L02?
	val03 = L02.cross(v03);		// test p3: left, right, or on top of line L02?
	if(val01*val01 < 10E-20 && val03*val03 < 10E-20)	
		// if BOTH p1 & p3 are on top of (collinear with) L02, 
		return(JTQUAD_DEGEN);	// then all points are collinear. zero area.
	if(val01*val01 < 10E-20)	// if p1 is collinear with line L02,
	{								// then L02 x V03 > 0 means
		if(val03 > 0.0) return(JTQUAD_TRI_012);	// a front-facing triangle; 
			else return(-JTQUAD_TRI_012);		// else backfacing.
	}
	if(val03*val03 < 10E-20)	// if p3 is collinear with L02,
	{								// then L02 x V01 > 0 means
		if(val01 > 0.0) return(-JTQUAD_TRI_230);	//backfacing triangle;
			else return(JTQUAD_TRI_230);			// front facing.
	}
	// at this point, we know that NEITHER p1 and p3 are on line L02. Are both
	// p1 and p3 on the SAME side of L02, or different sides? The 'val01' and
	// val03 values will have different signs if they're on different sides 
	// of L02, and same signs if on same side, so multiply:
	if(val01*val03 < 0.0) c02 = 2;	// TRUE if p1,p3 straddles line L02
	else c02 = 0;					// FALSE if they're on the same side of L02

	// Do similar tests for p0 and p2 for line L13:
	L13.sub(p3,p1);				// make the L13 vector from p1 to p3.
	v10.sub(p0,p1);				// make quad-side vectors from p1
	v12.sub(p2,p1);
	val10 = L13.cross(v10);		// test p0; left, right, or on top of line L13?
	val12 = L13.cross(v12);		// test p2; left, right, or on top of line L13?
	if(val10*val10 < 10E-20)	// if p0 is collinear with line L13,
	{								// then L13 x V12 > 0 means
		if(val12 > 0.0)	return(-JTQUAD_TRI_301);	// a backfacing triangle;
			else return(JTQUAD_TRI_301);			// else front-facing.
	}
	if(val12*val12 < 10E-20)	// if p2 is collinear with line L13,
	{								// then L13 x V10 > 0 means
		if(val10 > 0.0) return(JTQUAD_TRI_123);		// a front-facing triangle;
			else return(-JTQUAD_TRI_123);			// else back-facing.
	}
	// at this point, we know that NEITHER p0 or p2 are on line L13 either.
	// Are both p0 and p2 on the SAME side of L13, or different sides? The
	//	'val12' and 'val10' values will have different signs if they're on
	// different sides of L13, and same signs if on same side, so multiply:
	if(val10*val12 < 0.0) c13 = 1;	// TRUE if p0,p2 straddles line L13;
	else c13 = 0;					// FALSE if they're on same side of L13.
	
	// the quad is not degenerate or any kind of triangle; use c02,c13 to
	// decide the rest:
	switch(c02 + c13)
	{
	case 3:		// c02 true, c13 true
		if(v12.cross(L13) > 0.0)
			 return(JTQUAD_CONVEX);		// convex, front-facing
		else return(-JTQUAD_CONVEX);	// convex, back-facing
		break;
	case 2:		// c02 true, c13 false
		if(v12.cross(v10) > 0.0)
			 return(JTQUAD_CHEVRON_13); // chevron, front-facing,acute at p1,p3
		else return(-JTQUAD_CHEVRON_13);// chevron, back-facing.
		break;
	case 1:		// c02 false, c13 true
		if(v01.cross(v03) > 0.0)
			 return(JTQUAD_CHEVRON_02);// chevron, front-facing,acute at p0,p2
		else return(-JTQUAD_CHEVRON_02);
		break;
	case 0:		// c02 false, c13 false.
		return(JTQUAD_BOWTIE);		// bow-tie; front/back facing is ambiguous.
		break;
	default:
		printf("\n!?!? Vec2::ClassifyQuad() wierdness!!\n\n");
		return(JTQUAD_DEGEN);
		break;
	}
}

void Vec2::testClassifyQuad(void)
//------------------------------------------------------------------------------
// Simple examples to test quad-classify:
{
#define JT_PTMAX 9

Vec2 pt[9];

int i,j,m,n,result;

	pt[0].put(0.0, 0.0);		// make 4 corners of a unit square.
	pt[1].put(1.0, 0.0);
	pt[2].put(1.0, 1.0);
	pt[3].put(0.0, 1.0);

	pt[4].put(0.5, 0.5);		// and center point

	pt[5].put(0.2, 0.2);		// and smaller square nested inside unit square.
	pt[6].put(0.8, 0.2);
	pt[7].put(0.8, 0.8);
	pt[8].put(0.2, 0.8);

	for(i=0; i<JT_PTMAX; i++)			// loop through all possibilities
	{
		for(j=0; j<JT_PTMAX; j++)
		{
			for(m=0; m<JT_PTMAX; m++)
			{
				for(n=0; n<JT_PTMAX; n++)
				{
					if(i==j || i==m || i==n ||	// but don't duplicate vertices
					   j==m || j==n || 
					   m==n)	continue;
					result = classifyQuad(pt[i],pt[j],pt[m],pt[n]);
					switch(i)
					{					// print out the vertex names, in order
					case 0: printf("A  "); break;
					case 1: printf("B  "); break;
					case 2: printf("C  "); break;
					case 3: printf("D  "); break;
					case 4: printf("+  "); break;
					case 5: printf("a  "); break;
					case 6: printf("b  "); break;
					case 7: printf("c  "); break;
					case 8: printf("d  "); break;
					}
					switch(j)
					{
					case 0: printf("A  "); break;
					case 1: printf("B  "); break;
					case 2: printf("C  "); break;
					case 3: printf("D  "); break;
					case 4: printf("+  "); break;
					case 5: printf("a  "); break;
					case 6: printf("b  "); break;
					case 7: printf("c  "); break;
					case 8: printf("d  "); break;
					}
					switch(m)
					{
					case 0: printf("A  "); break;
					case 1: printf("B  "); break;
					case 2: printf("C  "); break;
					case 3: printf("D  "); break;
					case 4: printf("+  "); break;
					case 5: printf("a  "); break;
					case 6: printf("b  "); break;
					case 7: printf("c  "); break;
					case 8: printf("d  "); break;
					}
					switch(n)
					{
					case 0: printf("A  "); break;
					case 1: printf("B  "); break;
					case 2: printf("C  "); break;
					case 3: printf("D  "); break;
					case 4: printf("+  "); break;
					case 5: printf("a  "); break;
					case 6: printf("b  "); break;
					case 7: printf("c  "); break;
					case 8: printf("d  "); break;
					}
					printf(" result: %d\n",result);
				}
			}
		}
	}

#undef JT_PTMAX //locals only
}
//==============================================================================
//========================                ======================================
//========================   2x2 Matrices ======================================
//========================                ======================================
//==============================================================================

Mat2::Mat2(void)
//------------------------------------------------------------------------------
// void constructor
{
    zero();							// fill with zeroes.
}

void Mat2::zero(void) 
//------------------------------------------------------------------------------
// clear matrix to zero.
{
int i,j;
    
    for(i=0; i<VV; i++) 
	{
		for(j=0; j<VV; j++) 
		{
		    colrow[i][j] = (VECTYPE)0.0;
		}
	}
}

void Mat2::ident(void) 
//------------------------------------------------------------------------------
// set to identity matrix.
{
int i;

    zero();							// apply Mat2::zero() to the
									// current instance of Mat2;
    for(i=0; i< VV; i++) 
	{
		colrow[i][i] = (VECTYPE)1.0;			// put 1.0 on the diagonal
	}
}

void Mat2::scale(Mat2& in1, VECTYPE scale)
//------------------------------------------------------------------------------
// Make a scaled version of the matrix 'in1'.
{
	in1.colrow[0][0] *= scale;
	in1.colrow[0][1] *= scale;
	in1.colrow[1][0] *= scale;
	in1.colrow[1][1] *= scale;
}

void Mat2::prnt(void) 
//------------------------------------------------------------------------------
// prettyprint a matrix.
{
int i,j;

    dprintf("\n-------Mat2---------\n");
    for(j=0; j< VV; j++) 
	{
        for(i=0; i< VV; i++) 
		{
			dprintf("%f\t",colrow[i][j]);
	    }
		dprintf("\n");
	}
}

void Mat2::prnt(char *name)
//------------------------------------------------------------------------------
// prettyprint a LABELED matrix.
{
int i,j;

    dprintf("\n-------Mat2:%s----------\n",name);
    for(j=0; j< VV; j++) 
	{
        for(i=0; i< VV; i++) 
		{
			dprintf("%f\t",colrow[i][j]);
	    }
		dprintf("\n");
	}
}

void Mat2::mpy(Mat2& left, Mat2& rt) 
//------------------------------------------------------------------------------
//  Matrix-Matrix mpy: [th i s]   [l e ft][r  t ]
//					   [ -   -] = [-   - ][-  - ]				
{
int i,j,h;
Mat2 tmp;				

   
    for(j=0; j<VV; j++) 
	{
		for(i=0; i<VV; i++) 
		{
	    	tmp.colrow[i][j] = (VECTYPE)0.0;
	    	for(h=0; h<VV; h++) 
			{
				tmp.colrow[i][j] += 
				left.colrow[h][j] * rt.colrow[i][h];
			}
	    }
	}
    *this = tmp;					// copy result to 'self'
}


void Mat2::transpose(void)
//------------------------------------------------------------------------------
//   Replace 'self' with it's transpose.
{
int i,j;
VECTYPE tmp;

    for(i=0; i<VV; i++) 
	{
		for(j=0; j<i; j++) 
		{		// swap anti-diagonals.		
	    	tmp = colrow[i][j];
	    	colrow[i][j] = colrow[j][i];
	    	colrow[j][i] = tmp;
	    }
	}
}

void Mat2::makeRot270()
//------------------------------------------------------------------------------
//  Make a matrix that performs a 270-degree CCW rotation with no skew.
{
    colrow[0][0] =  (VECTYPE)0.0; 		// then overwrite with sines,cosines
    colrow[1][0] =  (VECTYPE)1.0; 
    colrow[0][1] = -(VECTYPE)1.0;
    colrow[1][1] =  (VECTYPE)0.0;
}

void Mat2::copyTo(Matrx& out)
//------------------------------------------------------------------------------
// Copy contents to 'Matrx' class object 'out'.  Resize 'out' as needed.
{
int i,j;

#if(PESKYWARNINGS==TRUE)
	if((out.getXsize()!=VV) || (out.getYsize()!=VV))
		DPF("!Mat2::copyTo() resized the destination matrix!");
#endif

	out.sizer(VV,VV);
	for(j=0; j<VV; j++)
	{
		for(i=0; i<VV; i++)
		{
			out.put0(j,i,colrow[i][j]);
		}
	}
}

//==============================================================================
//========================                ======================================
//========================   3x1 Vectors  ======================================
//========================                ======================================
//==============================================================================
// All member functions use internal tmp. storage, you can use the'this'
// keyword as a function argument without problems.

#undef VV
#define VV 3			// 3 element vectors, 3x3 element matrix

void Vec3::prnt(void)
//------------------------------------------------------------------------------
// Unlabeled prettyprint of Vec3 elements.
{
    dprintf("\nvec3:\t");
	dprintf("%f\t%f\t%f\n",row[0],row[1],row[2]);
}

void Vec3::prnt(char *name)
//------------------------------------------------------------------------------
// Labeled prettyprint of Vec3 elements.
{

	dprintf("\n%s: %f\t%f\t%f\n",name,row[0],row[1],row[2]);
}

VECTYPE Vec3::norm(Vec3& src)
//------------------------------------------------------------------------------
//Replace 'self' with unit-length version of 'src' vector; retn length.
{
int i;
VECTYPE length=(VECTYPE)0.0;
VECTYPE tmp=(VECTYPE)0.0;

    for(i=0; i< VV; i++) 
	{								// get vector length.
		length += src.row[i] * src.row[i];
	}
    if(length==(VECTYPE)1.0) 
	{
		*this = src;
		return((VECTYPE)1.0);			// already normalized.
	}

#if(IS_FLOAT==TRUE)
    length = sqrtf(length);
#endif

#if(IS_DOUBLE==TRUE)
    length = sqrt(length);
#endif

    if(length != (VECTYPE)0.0) 
	{									// normalize it.	
		tmp = (VECTYPE)1.0/ length;
		for(i=0; i< VV; i++) 
		{
	    	row[i] = tmp * src.row[i];
	    }
		return(length);
	}
    else 
	{
#if(DIAGNOSTICS==TRUE)
		DPF("\n ???Vec3::norm() given 0-length src vector!!!");
#endif
		return((VECTYPE)0.0);
	}
}

void Vec3::add(Vec3& src1, Vec3& src2) 
//------------------------------------------------------------------------------
// Vector addition: [this]= src1 + src2
{
int i;
Vec3 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i] + src2.row[i];
	}	
    *this = out;	
}

void Vec3::sub(Vec3& src1, Vec3& src2)
//------------------------------------------------------------------------------
// subtract vectors: [this] = src1 - src2
{
int i;
Vec3 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i] - src2.row[i];
	}
    *this = out;
}

void Vec3::scale(Vec3& src1,VECTYPE c)
//------------------------------------------------------------------------------
// Multiply by scalar: [this]= src1 + src2
{
int i;
Vec3 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i]*c;
	}
    *this = out;	
}

VECTYPE Vec3::dot(Vec3& src1, Vec3& src2)
//------------------------------------------------------------------------------
// vector dot product: [this] = src1 <dot> src2 	
{
VECTYPE out=(VECTYPE)0.0;
int i;

    for(i=0; i <VV; i++) 
		{
		out += src1.row[i] * src2.row[i];
		}
    return(out);    
}

void Vec3::cross(Vec3& src1, Vec3& src2) 
//------------------------------------------------------------------------------
// vector cross product: [this]= [in1]X[in2
{
Vec3 out;							// local store for result;

    out.row[0] = src1.row[1] * src2.row[2] - src1.row[2] * src2.row[1];
    out.row[1] = src1.row[2] * src2.row[0] - src1.row[0] * src2.row[2];
    out.row[2] = src1.row[0] * src2.row[1] - src1.row[1] * src2.row[0];
    *this = out;					// copy result to 'self'.
}

void Vec3::mpy(Mat3& srcM, Vec3& srcV)
//-----------------------------------------------------------------------------
//  Matrix-Vector multiply: [this]=[srcM][srcV]	
{
int i,j;
Vec3 dest;							// local store for result;

    for(j=0; j<VV; j++) 
	{
		for(i=0; i<VV; i++) 
		{
		    dest.row[j] += srcM.colrow[i][j] * srcV.row[i];
		}
	}
    *this = dest;	 				// copy results to calling Vec3
}

//==============================================================================
//========================                ======================================
//========================   3x3 Matrices ======================================
//========================                ======================================
//==============================================================================

void Mat3::zero(void)
//------------------------------------------------------------------------------
 // clear matrix to zero.
{
int i,j;
    
    for(i=0; i<VV; i++) 
	{
		for(j=0; j<VV; j++) 
		{
		    colrow[i][j] = (VECTYPE)0.0;
		}
	}
}

void Mat3::ident(void)
//------------------------------------------------------------------------------
// set to identity matrix.
{
int i;

    zero();								// apply Mat3::zero() to the
										// current instance of Mat3;
    for(i=0; i< VV; i++) 
	{
		colrow[i][i] = (VECTYPE)1.0;	// put 1.0 on the diagonal
	}
}

void Mat3::copy(Mat3& src)
//------------------------------------------------------------------------------
{
int i,j;

	for(i=0; i<VV; i++) 
	{
		for(j=0; j<VV; j++) 
		{
			colrow[i][j] = src.colrow[i][j];
		}
	}
}

void Mat3::prnt(void)
//------------------------------------------------------------------------------
// prettyprint a matrix.
{
int i,j;

    DPF("\n-------Mat3---------\n");
    for(j=0; j< VV; j++) 
	{
        for(i=0; i< VV; i++) 
		{
			dprintf("%f\t",colrow[i][j]);
		}
		dprintf("\n");
	}
}

void Mat3::prnt(char *name)
//------------------------------------------------------------------------------
// prettyprint a LABELED matrix.
{
int i,j;

    dprintf("\n-------Mat3:%s----------\n",name);
    for(j=0; j< VV; j++) 
	{
        for(i=0; i< VV; i++) 
		{
			dprintf("%f\t",colrow[i][j]);
	    }
		dprintf("\n");
	}
}

void Mat3::mpy(Mat3& left, Mat3& rt)
//------------------------------------------------------------------------------
//  Matrix-Matrix mpy: [th i s]   [l e ft][r t -]
//					   [ - - -] = [- -  -][- - -]
//					   [ - - -]   [- -  -][- - -]					 
{
int i,j,h;
Mat3 tmp;				

   
    for(j=0; j<VV; j++) 
	{
		for(i=0; i<VV; i++) 
		{
			tmp.colrow[i][j] = (VECTYPE)0.0;
	    	for(h=0; h<VV; h++) 
			{
				tmp.colrow[i][j] += 
				left.colrow[h][j] * rt.colrow[i][h];
			}
	    }
	}
    *this = tmp;					// copy result to 'self'
}

void Mat3::convolve(Mat2& inA2, Mat2& inB2)
//------------------------------------------------------------------------------
// inA, inB are coefficient matrices of two polynomials; find the 
// the product of these two polynomials by convolving their coef. matrices.
{
int i,j;
VECTYPE wt;

	zero();							// clear out current contents.
	for(i=0; i<2; i++)				// for each element in inA,
	{
		for(j=0; j<2; j++)
		{
			wt = inA2.colrow[i][j];	// accumulate a shifted, scaled copy of inB.
			colrow[i  ][j  ] += wt * inB2.colrow[0][0];
			colrow[i  ][j+1] += wt * inB2.colrow[0][1];
			colrow[i+1][j  ] += wt * inB2.colrow[1][0];
			colrow[i+1][j+1] += wt * inB2.colrow[1][1];
		}
	}
}

void Mat3::transpose(void)
//------------------------------------------------------------------------------
//   Replace 'self' with it's transpose.
{
int i,j;
VECTYPE tmp;

    for(i=0; i<VV; i++) 
	{
		for(j=0; j<i; j++) 
		{							// swap anti-diagonals.		
	    	tmp = colrow[i][j];
	    	colrow[i][j] = colrow[j][i];
	    	colrow[j][i] = tmp;
	    }
	}
}

void Mat3::copyTo(Matrx& out)
//------------------------------------------------------------------------------
// Copy contents to 'Matrx' class object 'out'.  Resize 'out' as needed.
{
int i,j;

#if(PESKYWARNINGS==TRUE)
	if((out.getXsize()!=VV) || (out.getYsize()!=VV))
		DPF("!Mat3::copyTo() resized the destination matrix!");
#endif

	out.sizer(VV,VV);
	for(j=0; j<VV; j++)
	{
		for(i=0; i<VV; i++)
		{
			out.put0(j,i,colrow[i][j]);
		}
	}
}


//==============================================================================
//========================                ======================================
//========================   4x1 Vectors  ======================================
//========================                ======================================
//==============================================================================

#undef VV
#define VV 4			// 4 element vectors, 4x4 element matrix

void Vec4::prnt(void) 
//------------------------------------------------------------------------------
// Unlabeled prettyprint of Vec3 elements.
{
    dprintf("\nvec4:\t %f,/t%f,/t%f/t%f/n",row[0],row[1],row[2],row[3]);

}

void Vec4::prnt(char *name) 
//------------------------------------------------------------------------------
// Labeled prettyprint of Vec4 elements.
{
	dprintf("\n%s:\t%f\t%f\t%f\t%f\n",name,row[0],row[1],row[2],row[3]);
}

void Vec4::copyTo(Matrx &out)		
//------------------------------------------------------------------------------
// resize the 'out' matrix to a 4x1 column vector, and copy contents to it.
{
int i;

	out.sizer(4,1);
	for(i=0; i<4; i++)  out.put0(i,0,row[i]);
}

VECTYPE Vec4::norm(Vec4& src)
//------------------------------------------------------------------------------
//Replace 'self' with unit-length version of 'src' vector; return length.
{
int i;
VECTYPE length = (VECTYPE)0.0;
VECTYPE tmp    = (VECTYPE)0.0;

    for(i=0; i< VV; i++) 
	{									// get vector length.
		length += src.row[i] * src.row[i];
	}
    if(length==(VECTYPE)1.0) 
	{
		*this = src;
		return((VECTYPE)1.0);			// already normalized.
	}

#if(IS_FLOAT==TRUE)
    length = sqrtf(length);
#endif

#if(IS_DOUBLE==TRUE)
    length = sqrt(length);
    DPF("Vec4:doublenorm");
#endif

    if(length != (VECTYPE)0.0) 
	{								// normalize it.
		tmp = (VECTYPE)1.0/ length;
		for(i=0; i< VV; i++) 
		{
	 		row[i] = tmp * src.row[i];
		}
		return(length);
	}
    else 
	{
#if(DIAGNOSTICS==TRUE)
		DPF("\n ???Vec4::norm() given 0-length src vector!!!");
#endif
		return((VECTYPE)0.0);
	}
}

void Vec4::add(Vec4& src1, Vec4& src2) 
//------------------------------------------------------------------------------
// Vector addition: [this]= src1 + src2
{
int i;
Vec4 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i] + src2.row[i];
	}
    *this = out;	
}

void Vec4::sub(Vec4& src1, Vec4& src2) 
//------------------------------------------------------------------------------
// subtract vectors: [this] = src1 - src2
{
int i;
Vec4 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i] - src2.row[i];
	}
    *this = out;
}

void Vec4::scale(Vec4& src1,VECTYPE c) 
//------------------------------------------------------------------------------
// Multiply by scalar: [this]= src1 + src2
{
int i;
Vec4 out;

    for(i=0; i<VV; i++) 
	{
		out.row[i] = src1.row[i]*c;
	}
    *this = out;	
}

VECTYPE Vec4::dot(Vec4& src1, Vec4& src2)
//------------------------------------------------------------------------------
// vector dot product: [this] = src1 <dot> src2 	
{
VECTYPE out=(VECTYPE)0.0;
int i;

    for(i=0; i <VV; i++) 
	{
		out += src1.row[i] * src2.row[i];
	}
    return(out);    
}

void Vec4::cross(Vec4& src1, Vec4& src2) 
//------------------------------------------------------------------------------
// Cross product of the 1st 3 terms of two 4-element vectors that have 
// been normalized by their 4th element (i.e. homogeneous coords used).
{
VECTYPE norm;
Vec4 out;

    if(src1.row[3] == (VECTYPE)0.0) 
	{
		DPF("\n!!!Vec4::cross(): src1.row[3]=0.0, so NOP!!!\n");
		return;
	}
    if(src2.row[3] == (VECTYPE)0.0) 
	{
		DPF("\n!!!Vec4::cross(): src2->row[3]=0.0, so NOP!!!\n");
		return;
	}
    if(src1.row[3] != (VECTYPE)1.0) 
	{
#if(DIAGNOSTICS==TRUE)
		DPF("\n!!WARNING!! Vec4::cross(): src1 norm'd by src1.row[3]!!!\n");
#endif
		norm = (VECTYPE)1.0 / src1.row[3];
		src1.row[0] *= norm;
		src1.row[1] *= norm;
		src1.row[2] *= norm;
		src1.row[4] = (VECTYPE)1.0;
	}
    if(src2.row[3] != (VECTYPE)1.0) 
	{
#if(DIAGNOSTICS==TRUE)
		DPF("\n!!WARNING!! Vec4::cross(): src2 norm'd by src2.row[3]!!!\n");
#endif
		norm = (VECTYPE)1.0 / src2.row[3];
		src2.row[0] *= norm;
		src2.row[1] *= norm;
		src2.row[2] *= norm;
		src2.row[3] = (VECTYPE)1.0;
	}
    out.row[0] = src1.row[1]* src2.row[2] - src1.row[2]* src2.row[1];
    out.row[1] = src1.row[2]* src2.row[0] - src1.row[0]* src2.row[2];
    out.row[2] = src1.row[0]* src2.row[1] - src1.row[1]* src2.row[0];
    out.row[3] = (VECTYPE)1.0;
    *this = out;
}

void Vec4::mpy(Mat4& srcM, Vec4& srcV)
//----------------------------------------------------------------------
//  Matrix-Vector multiply: [this]=[srcM][srcV]	
{
int i,j;
Vec4 dest;							// local store for result;

    for(j=0; j<VV; j++) 
	{
		for(i=0; i<VV; i++) 
		{
		    dest.row[j] += srcM.colrow[i][j] * srcV.row[i];
		}
	}
    *this = dest;					// copy results to calling Vec4
}

//==============================================================================
//========================                ======================================
//========================   4x4 Matrices ======================================
//========================                ======================================
//==============================================================================

void Mat4::zero(void) 
//------------------------------------------------------------------------------
// clear matrix to zero.
{
int i,j;
    
    for(i=0; i<VV; i++) 
	{
		for(j=0; j<VV; j++) 
		{
	 		colrow[i][j] = (VECTYPE)0.0;
	 	}
	}
}

void Mat4::ident(void)
//------------------------------------------------------------------------------
// set to identity matrix.
{
int i;

    zero();								// apply Mat4::zero() to the
										// current instance of Mat4;
    for(i=0; i< VV; i++) 
	{
		colrow[i][i] = (VECTYPE)1.0;	// put 1.0 on the diagonal
	}
}

void Mat4::prnt(void)
//------------------------------------------------------------------------------
// prettyprint a matrix.
{
int i,j;


    dprintf("\n----------Mat4------------\n");
    for(j=0; j< VV; j++) 
	{
		for(i=0; i< VV; i++) 
		{
	    	dprintf("%f\t",colrow[i][j]);
	    }
		dprintf("\n");
	}
}

void Mat4::prnt(char *name)
//------------------------------------------------------------------------------
// prettyprint a LABELED matrix.
{
int i,j;


    dprintf("\n---------Mat4:%s------------\n",name);
    for(j=0; j< VV; j++) 
	{
		for(i=0; i< VV; i++) 
		{
			dprintf("%f\t",colrow[i][j]);
	    }
		dprintf("\n");
	}
}

void Mat4::mpy(Mat4& left, Mat4& rt) 
//------------------------------------------------------------------------------
//  Matrix-Matrix mpy: 			[t h i s]   [l e f t][r t - -]
//								[- - - -] = [- - - -][- - - -]
//								[- - - -]   [- - - -][- - - -]
//								[- - - -]   [- - - -][- - - -]		  
{
int i,j,h;
Mat4 tmp;				

    for(j=0; j<VV; j++) 
	{
		for(i=0; i<VV; i++) 
		{
			tmp.colrow[i][j]=(VECTYPE)0.0;
			for(h=0; h<VV; h++) 
			{
				tmp.colrow[i][j] += 
				left.colrow[h][j] * rt.colrow[i][h];
			}
	    }
	}
    *this = tmp;						// copy result to 'self'
}

void Mat4::convolve(Mat2& inA2, Mat3& inB3)
//------------------------------------------------------------------------------
// inA, inB are coefficient matrices of two polynomials; find the 
// the product of these two polynomials by convolving their coef. matrices.
{
int i,j;
VECTYPE wt;

	zero();							// clear out current contents.
	for(i=0; i<3; i++)				// for each element of inB,
	{
		for(j=0; j<3; j++)
		{
			wt = inB3.colrow[i][j];	// accumulate a shifted, scaled copy of inA
			colrow[i  ][j  ] += wt * inA2.colrow[0][0];
			colrow[i  ][j+1] += wt * inA2.colrow[0][1];
			colrow[i+1][j  ] += wt * inA2.colrow[1][0];
			colrow[i+1][j+1] += wt * inA2.colrow[1][1];
		}
	}
}

void Mat4::transpose(void)
//------------------------------------------------------------------------------
//   Replace 'self' with it's transpose.
{
int i,j;
VECTYPE tmp;

    for(i=0; i<VV; i++) 
		{
		for(j=0; j<i; j++) 
		{								// swap anti-diagonals.		
	    	tmp = colrow[i][j];
	    	colrow[i][j] = colrow[j][i];
	    	colrow[j][i] = tmp;
	    }
	}

}

void Mat4::copyTo(Matrx& out)
//------------------------------------------------------------------------------
// Copy contents to 'Matrx' class object 'out'.  Resize 'out' as needed.
{
int i,j;

#if(PESKYWARNINGS==TRUE)
	if((out.getXsize()!=VV) || (out.getYsize()!=VV))
		DPF("!Mat4::copyTo() resized the destination matrix!");
#endif

	out.sizer(VV,VV);
	for(j=0; j<VV; j++)
	{
		for(i=0; i<VV; i++)
		{
			out.put0(j,i,colrow[i][j]);
		}
	}
}

//==============================================================================
//========================                ======================================
//========================   NxM Matrices ======================================
//========================                ======================================
//==============================================================================

Matrx::Matrx(void)
//------------------------------------------------------------------------------
// Constructor for an 'empty' matrix object.
{
    a = NULL;					// make an empty array for pixels.
    xsize = 0; 
	ysize = 0;
}


Matrx::~Matrx(void)
//------------------------------------------------------------------------------
// destructor; de-alloc memory before we destroy the object.
{
    if(a != NULL) 
	{
    	delete[xsize*ysize] a;		// release memory, if any
    	a = NULL;					// (for VC++)
    } 
#if (PESKYWARNINGS==TRUE)
    else 
	{
		dprintf("\n!!empty matrix called destructor!!\n");
	}
#endif 
    xsize = 0;						// set size to zero.
	ysize = 0;
}

BOOL Matrx::wipe(void)
//------------------------------------------------------------------------------
// release space for matrix; make our size zero. If already zero sized,
// return FALSE, else return TRUE.
{
    if(a==NULL)
	{								// array is already empty;
#if(PESKYWARNINGS==TRUE)
    	dprintf("\n!!Matrx::wipe called on empty object!!\n");
#endif
		xsize = 0;					// be sure its other members are
									// also benign.
		ysize = 0;
		return(FALSE);	
	}
    delete[xsize*ysize] a;			// release all pixel memory.
	a = NULL;						// (for VC++)
    xsize = 0;
	ysize = 0;
    return(TRUE);
}

void Matrx::sizer(int iysize,int ixsize)
//------------------------------------------------------------------------------
// allocate space for matrix.
{
    if(a != NULL) 
	{								// do we have ANY entries at all?
#if(PESKYWARNINGS==TRUE)
    	dprintf("\n!!Matrx::sizer called on nonempty object!!\n");
#endif 
									// YES. correct # of entries?
		if(xsize != ixsize || ysize != iysize) 
		{
			wipe();					// NO discard all existing entries;
		}
		else 
		{
			return;					// YES. no resizing needed!
		}
	}
    a = new VECTYPE[ixsize*iysize];	//make array of size 'xsize*ysize'
#if(DIAGNOSTICS==TRUE)
    if(a==NULL) 
	{
		dprintf("\n Matrx::sizer() unable to allocate memory!\n");
		exit(1);
	}
#endif
    xsize = ixsize;
	ysize = iysize;
}

void Matrx::sizer(Matrx& src){	
//------------------------------------------------------------------------------
// Allocate memory for the same number of elements found in 'src'.

	if(src.a==NULL || src.xsize==0 || src.ysize==0) 
	{
		dprintf("\n!!Matrx::sizer called with empty src object!\n");
		exit(1);
	}
    if(a != NULL) 
	{								// do we have ANY pixels at all?
#if(PESKYWARNINGS==TRUE)			// YES;  warn user.
    	dprintf("\n!!Matrx::sizer called on nonempty object!!\n");
#endif 
		if(xsize != src.xsize || ysize != src.ysize) 
		{							// do we have the correct # of pix?
			wipe();					// NO-- so discard all & start over:
		}
		else 
		{						// yes-- no resizing needed!
			return;
		}
	}
    a = new VECTYPE[src.xsize * src.ysize];	// make a pixel array/
#if(DIAGNOSTICS==TRUE)
    if(a==NULL) 
	{
		dprintf("\n Matrx::sizer() unable to allocate memory!\n");
		exit(1);
	}
#endif
    xsize = src.xsize;
	ysize = src.ysize;
}

BOOL Matrx::copy(Matrx& src)
//------------------------------------------------------------------------------
// auto-resizing entry-by-entry copy (no change to our object's name). 
// If size of 'src' is different from us, 'wipe' our current contents,
//  and change our size to match 'src' before copy.
// If a 'wipe' was necessary, return 'FALSE', else return 'TRUE'.
{
BOOL out=TRUE;

    if(xsize != src.xsize || ysize != src.ysize)
	{
    								// src size != us, so wipe
		wipe();
		sizer(src);
		out=FALSE;
	}
	memcpy(a, src.a, xsize * ysize * sizeof(VECTYPE));
    return(out);
}

void Matrx::put0(int row,int col, VECTYPE val) 
//------------------------------------------------------------------------------
// write 'val' from spec'd row, col. using zero-based indexing.
{	
#if(PESKYWARNINGS==TRUE)
	ASSERT(row>=0 && row<ysize);
	ASSERT(col>=0 && col<xsize);
#endif
		a[row + col*ysize] = val;			
}

void Matrx::put1(int row, int col, VECTYPE val) 
//------------------------------------------------------------------------------
// read 'val' from spec'd row, col.
{
#if(PESKYWARNINGS==TRUE)
	ASSERT(row>0 && row<=ysize);
	ASSERT(col>0 && col<=xsize);
#endif
		a[(row-1) + (col-1)*ysize] = val;
}

void Matrx::copy0Col(int destCol, Matrx &src, int srcCol)
//------------------------------------------------------------------------------
// Copy one column from 'src' matrix.  Column 'srcCol' of matrix 'src' is copied
// element-by-element to column 'destCol' in 'this'. 
// CAREFUL! DOES NOT CHECK ALL MATRIX AND VECTOR SIZES!
{
#if(PESKYWARNINGS==TRUE)
	ASSERT(destCol>=0 && destCol<xsize);		// out-of-range column number!
	ASSERT(srcCol>=0 && srcCol<src.getXsize());	// out-of-range column number!
	ASSERT(src.getYsize()==getYsize());			// column size mismatch!!
#endif

int i,imax;
	
	imax = src.getYsize();
	if(imax > getYsize())						// is src column size > dest?
	{
		imax = getYsize();						// truncate, and
		ASSERT(FALSE);							// complain.
	}
	for(i=0; i<imax; i++)				// do element-by-element copy.
	{
		put0(i,destCol,src.get0(i,srcCol));
	}
}


VECTYPE Matrx::get0(int row, int col)
//------------------------------------------------------------------------------
// read 'val' from spec'd row, col.
{
#if(PESKYWARNINGS==TRUE)
	ASSERT(row>=0 && row<ysize);
	ASSERT(col>=0 && col<xsize);
#endif
	return(a[row + col*ysize]);
}

VECTYPE Matrx::get1(int row, int col)
//------------------------------------------------------------------------------
// read 'val' from spec'd row, col. using 1-based indexing
{
#if(PESKYWARNINGS==TRUE)
	ASSERT(row>0 && row<=ysize);
	ASSERT(col>0 && col<=xsize);
#endif
	return(a[(row-1) + (col-1)*ysize]);
}

BOOL Matrx::add(Matrx &left, Matrx& rt)
//------------------------------------------------------------------------------
// 'this' = left + right. Return FALSE if mismatched matrix sizes.
{
int i;
Matrx* ptmp;
BOOL useTmp;

	ptmp = this;					//  point to ourselves, the output matrix;
	useTmp = FALSE;					//  and assume we'll compute answers there.
	if(&left==this || &rt==this)	// but if output overwrites an input,
	{
		ptmp = new Matrx;			// make a temporary output matrix;
		ptmp->sizer(rt);	
		useTmp=TRUE;				// and set flag so we copy it to output.
	}			
									// check for size mismatches.
#if(DIAGNOSTICS==TRUE)
	ASSERT(xsize == left.getXsize() && xsize == rt.getXsize()); //ensure size match
	ASSERT(ysize == left.getYsize() && ysize == rt.getYsize()); 
#endif

	for(i=0; i<xsize*ysize; i++)
	{
		ptmp->a[i] = left.a[i] + rt.a[i];
	}
	if(useTmp==TRUE)				// if answer is in temporary storage,
	{
		copy(*ptmp); 						// replace 'this' with temp.
		ptmp->wipe();						// and discard temp. version.
		delete ptmp;
	}
	return(TRUE);
}

BOOL Matrx::sub(Matrx &left, Matrx& rt)
//------------------------------------------------------------------------------
// 'this' = left - right. Return FALSE if mismatched matrix sizes.
{
int i;
Matrx* ptmp;
BOOL useTmp;

									// check for size mismatches.
#if(DIAGNOSTICS==TRUE)
	ASSERT(xsize != left.getXsize() || xsize != rt.getXsize()); //size mismatch!
	ASSERT(ysize != left.getYsize() || ysize != rt.getYsize()); 
#endif

	ptmp = this;					//  point to ourselves, the output matrix;
	useTmp = FALSE;					//  and assume we'll compute answers there.
	if(&left==this || &rt==this)	// but if output overwrites an input,
	{
		ptmp = new Matrx;			// make a temporary output matrix;
		ptmp->sizer(rt);	
		useTmp=TRUE;				// and set flag so we copy it to output.
	}			
	for(i=0; i<xsize*ysize; i++)
	{
		ptmp->a[i] = left.a[i] - rt.a[i];
	}
	if(useTmp==TRUE)				// if answer is in temporary storage,
	{
		copy(*ptmp); 						// replace 'this' with temp.
		ptmp->wipe();						// and discard temp. version.
		delete ptmp;
	}
	return(TRUE);
}

void Matrx::addTo0(int row, int col, VECTYPE val)
//------------------------------------------------------------------------------
// Add 'val' to the element currently stored at (row,col).  Return FALSE if
// element doesn't exist.
{
#if(PESKYWARNINGS==TRUE)
	ASSERT(row>=0 && row<ysize);
	ASSERT(col>=0 && col<xsize);
#endif
	a[row + col*ysize] += val;
}

void Matrx::addTo1(int row, int col, VECTYPE val)
//------------------------------------------------------------------------------
// Add 'val' to the element currently stored at (row,col).  Return FALSE if
// element doesn't exist.
{
#if(PESKYWARNINGS==TRUE)
	ASSERT(row>0 && row<=ysize);
	ASSERT(col>0 && col<=xsize);
#endif
	a[(row-1) + (col-1)*ysize] += val;
}

void Matrx::ident()
//------------------------------------------------------------------------------
// Make 'this' into an identity matrix without resizing.  'this' must be square.
{
int i,imax;

	imax = getXsize();
	ASSERT(getYsize()==imax);				// non-square matrix!!
	zero();									// clear all contents
	for(i=0; i<imax; i++)
	{
		put0(i,i,(VECTYPE)1.0);
	}
}
void Matrx::mpyScalar(VECTYPE wt)
//------------------------------------------------------------------------------
// Multiply all elements in matrix by scalar value 'wt'.
{
int i;

	for(i=0; i<xsize*ysize; i++) 
	{
		a[i] *= wt;
	}
}

BOOL Matrx::transpose(Matrx &src)
//------------------------------------------------------------------------------
{
int i,j,imax,jmax;
Matrx destmp;							// temp storage in case src==this.
Matrx *pDest;	

	imax = src.getXsize();
	jmax = src.getYsize();
#if(DIAGNOSTICS==TRUE)
	ASSERT(imax==jmax);					// be sure 'src' is square
#endif
	pDest = this;						// point to the output;
	if(this==&src)						// if output is also the input, then
	{
		pDest = &destmp;				// point to a temporary output buffer.
	}
	pDest->sizer(jmax,imax);			// Set output size using swapped x,ysize

    for(i=0; i<imax; i++) 
	{
		for(j=0; j<jmax; j++) 
		{									// Copy input to output, but	
	    	pDest->put0(j,i,get0(i,j));		// with swapped indices
	    }
	}
	return(TRUE);
}

BOOL Matrx::mmpy(Matrx& left, Matrx& rt)
//------------------------------------------------------------------------------
// Replace our current contents with result of matrix multiply [left][rt]. 
// Return FALSE on err.  
// Uses a 'temp' matrix to store intermediate results, so a matrix object can 
// use ITSELF as an argument to mmpy.  
// EXAMPLE;  Matrx us,them;		us.mmpy(us,them)
// (this is SLOWER but more flexible than requiring non-self inputs)
{
Matrx *ptmp;
int r,c,k,rmax,cmax,kmax;
VECTYPE val;
BOOL useTmp;						// true if output held in temp. storage

#if(DIAGNOSTICS==TRUE)
	if(left.getXsize() != rt.getYsize()) 
	{
		dprintf("Matrx::mmpy() size mismatch!!");
		return(FALSE);
	}
#endif
	cmax = rt.getXsize();			// #columns in  output== #columns in rt,
	rmax = left.getYsize();			// #rows in output==#rows in left.
	kmax = rt.getYsize();			// # elements in a col. of rt & row of left
	ptmp = this;					//  point to ourselves, the output matrix;
	useTmp = FALSE;					//  and assume we'll use it.
	if(&left==this || &rt==this)	// if output overwrites an input,
	{
		ptmp = new Matrx;			// make a temporary output matrix;
		useTmp=TRUE;				// and set flag so we copy it to output.
	}								
	if(ptmp->getXsize()!=cmax || ptmp->getYsize()!=rmax)
	{
		ptmp->sizer(rmax,cmax);		// resize as needed.
	}
	for(c=0; c<cmax; c++) {
		for(r=0; r<rmax; r++) {
			val = (VECTYPE)0.0;		// accumulate a row/col multiply
			for(k=0; k<kmax; k++) {
				val += left.get0(r,k) * rt.get0(k,c);
				}
			ptmp->put0(r,c,val);
			}
		}
	if(useTmp==TRUE)				// if answer is in temporary storage,
	{
		copy(*ptmp); 						// replace 'this' with temp.
		ptmp->wipe();						// and discard temp. version.
		delete ptmp;
	}
	return(TRUE);
}

void Matrx::randFill(VECTYPE min, VECTYPE max)
//------------------------------------------------------------------------------
// fill matrix with random #s.
{
int r,c;

	if(min<max) 
	{
		for(r=0; r<ysize; r++) 
		{									// for each row,
			for(c=0; c<xsize; c++)
			{
				put0(r,c,
				min + (max-min)*(VECTYPE)(rand()/(float)RAND_MAX));
			}
		}
	}
	else 
	{
		for(r=0; r<ysize; r++) 
		{									// for each row,
			for(c=0; c<xsize; c++)
			{
				put0(r,c,(VECTYPE)rand());
			}
		}
	}
}

void Matrx::zero(void)
//------------------------------------------------------------------------------
// set all pixels to 0.0
{
int r,c;

	for(r=0; r<ysize; r++) 
	{									// for each row,
		for(c=0; c<xsize; c++)
		{
			put0(r,c,(VECTYPE)0.0);
		}
	}
}

void Matrx::countR(void)
//------------------------------------------------------------------------------
// fill with row-major-order counting 
// (1 2 3)
// (4 5 6)
// (7 8 9);
{
int i,r,c;

	i=1;
	for(r=0; r<ysize; r++) 
	{									// for each row,
		for(c=0; c<xsize; c++)
		{
			put0(r,c,(VECTYPE)i);
			i++;
		}
	}
}

void Matrx::countC(void)
//------------------------------------------------------------------------------
// fill with column-major-order counting 
// (1 4	7)
// (2 5	8)
// (3 6	9)
{
int i,r,c;

	i=1;
	for(c=0; c<xsize; c++) 
	{										// for each column,
		for(r=0; r<ysize; r++) 
		{
			put0(r,c,(VECTYPE)i);
			i++;
		}
	}
}

BOOL Matrx::isEqual(Matrx &src,VECTYPE tol2)
//------------------------------------------------------------------------------
// Element-by-element, compare 'this' to 'src' matrix. Return TRUE iff all
// elements match (or differ by less than sqrt(tol2) 
// AND matrices are the same size.
{
int i,j,imax,jmax;
VECTYPE diff;

	if(getXsize()!=src.getXsize() || getYsize()!=src.getYsize())
	{
		return(FALSE);						// mismatched size!
	}
	imax = getXsize();
	jmax = getYsize();		
	for(i=0; i<imax; i++)
	{
		for(j=0; j<jmax; j++)
		{
			diff = src.get0(j,i) - get0(j,i);
			if(diff*diff > tol2) return(FALSE);
		}
	}
	return(TRUE);
}

void Matrx::copyTo(Mat2& out)
//------------------------------------------------------------------------------
// copy first 2x2 elements to 'out'.  Complain if there's a size mismatch
{
#if(DIAGNOSTICS==TRUE)
	ASSERT(getXsize()>=2 && getYsize()>=2);
#endif
	out.colrow[0][0] = get0(0,0);
	out.colrow[0][1] = get0(1,0);
	out.colrow[1][0] = get0(0,1);
	out.colrow[1][1] = get0(1,1);
}

void Matrx::copyTo(Mat3& out)
//------------------------------------------------------------------------------
// copy first 3x3 elements to 'out'.  Complain if there's a size mismatch
{
int i,j;
	
#if(DIAGNOSTICS==TRUE)
	ASSERT(getXsize()>=3 && getYsize()>=3);
#endif
	for(j=0; j<3; j++)
	{
		for(i=0; i<3; i++)
		{
			out.colrow[i][j] = get0(j,i);
		}
	}
}

void Matrx::copyTo(Mat4& out)
//------------------------------------------------------------------------------
// copy first 4x4 elements to 'out'.  Complain if there's a size mismatch
{
int i,j;

#if(DIAGNOSTICS==TRUE)
	ASSERT(getXsize()>=4 && getYsize()>=4);
#endif
	for(j=0; j<4; j++)
	{
		for(i=0; i<4; i++)
		{
			out.colrow[i][j] = get0(j,i);
		}
	}
}

void Matrx::textOut(FILE* fptr, char* title=NULL, int linewidth=80)
//------------------------------------------------------------------------------
// Pretty-print the contents of this matrix to the stream 'fptr' (usually
// a file).
{
int r,c;
char* msg;								// text message buffer
char nu[40];							// string holding 1 element.
int line;

#if(DIAGNOSTICS==TRUE)
	if(0==xsize || 0==ysize) 
	{
		dprintf("Matrx::textOut() given zero-sized matrix!");
		return;
	}
#endif	 
	
	if(title!=NULL) 
	{									// print a title,
		fprintf(fptr,
		"\n\n------------------%s--------------------------\n",title);
	}
	else
	{
		fprintf(fptr,
		"\n\n----------------------------------------------\n");
	}
	msg = new char[linewidth+1];		// make a line buffer.
	line=0;								// # of chars already printed
										// on this line.
	msg[0] = '\0';						// empty the line buffer.

	for(r=0; r<ysize; r++) 
	{									// for each row of matrix,
		for(c=0; c<xsize; c++) 
		{								// & each element in the row,
			sprintf(nu,"%f ",get0(r,c));// make an ASCII string;
			line += strlen(nu); 		// see if it fits this line;
			if(line < linewidth) 
			{							// YES! append the entry
				strcat(msg,nu);			// to our growing 'msg' buffer
										// and 'tab' to next position
				while(line%8!=0 && line<linewidth) 
				{
					strcat(msg," ");
					line++;
				}
			}
			else 
			{							// NO! print what we've got,
				fprintf(fptr,"%s\n     ",msg);	// go to next line,
				line=5;					// and indent by 5 spaces,
				msg[0] = '\0';			// clear the msg buffer,
				strcat(msg,nu);			// and add our new ASCII string
			}
		}
		fprintf(fptr,"%s",msg);			// end-of-row. print msg buffer.
		line=0;		  					// & clear it for the next line.
		msg[0] = '\0';
	}
	fprintf(fptr,"\n\n");				// double-space at end.
	delete msg;							// free line buffer memory.
}

void Matrx::prnt(const char *name)
//------------------------------------------------------------------------------
// Pretty-print the contents of this matrix using 'dprintf' macro defined at
// top of file (usually VC++/MFC's 'TRACE' macro).  'Linewidth' is max. allowed
// width of text line.
// --May 3,2001 J. Tumblin; revised to fix spurious buffer overflow problem
{
#define JT_WIDE		120					// # of chars allowed on one line.
int r,c;								// row, column.
char msg[JT_WIDE+1];					// text message buffer
char digits[JT_WIDE+1];					// string holding 1 numerical element.
char nu[JT_WIDE+1];						// Any req'd leading spaces+'digits[]
int line;								// width of current text line.
int col;								// width of current column
#if(DIAGNOSTICS==TRUE)
	if(0==xsize || 0==ysize) {
		dprintf("Matrx::prnt() given zero-sized matrix!");
		return;
		}
#endif	 
#define JT_TABSTOP	8					// column width in printouts.

	// !!!DON'T SPLIT THE NEXT LINE!!!! dprintf is a MACRO!!!
	dprintf("\n-----------%s: %d rows, %d cols----------\n",name,getYsize(),getXsize());
	for(r=0; r<ysize; r++)				// for each row of matrix,
	{
		msg[0] = '\0';					// empty the line buffer.
		line = 0;						// init.# of chars printed on this line.
		col = 0;						// init. # of chars in current column;
		for(c=0; c<xsize; c++)			// & each element in the row,
		{								// Make orderly printed columns:-------
			sprintf(digits,"%f ",get0(r,c));// Make a string of digits,
			nu[0] = '\0';				// (empty the column-holding buffer)
			col = strlen(digits);		// init column width to # of digits,
			if(get0(r,c)>=0.0 && col<(JT_WIDE-1))	
										// for non-negative#s, add a leading
			{							// space to the column buffer:
				strcat(nu," ");
				col++;					//(that's one more char for this column)
			}
			if(c!=0)					// For all but the 1st matrix column,
			{							// add leading spaces to make all column 
				while(col%JT_TABSTOP!=0 && col<(JT_WIDE-1))
				{						// widths an integer multiple of TABSTOP
					strcat(nu," ");	
					col++;	
				}
			}
			strcat(nu,digits);			// NOW put digits after leading spaces
										// to complete the column of text.
			if(line+col < (JT_WIDE-1))	// Will this new column fit on the line?
			{
				strcat(msg,nu);			// YES! Put column on this line.
				line += col;
			}
			else						// NO! print what we've got,
			{
				dprintf("%s\n     ",msg);// indent the next line by 5 spaces,
				line=5;					// shorten text on it by 5 chars,
				msg[0] = '\0';			// clear the line buffer, and
										// continue this long row of the
										// matrix; for 1st continued column,
				if(get0(r,c)>=(VECTYPE)0.0)	strcat(msg," ");
										// add a leading space for non-neg. #s,
				strcat(msg,digits);		// and then the digits of the element.
			}
		}
		dprintf("%s\n",msg);			// matrix end-of-row; print msg buffer.
		line=0;		  					// & clear it for the next line.
		msg[0] = '\0';
	}
//	dprintf("\n\n");					// double-space at end.
#undef JT_TABSTOP
#undef JT_WIDE
}

//==============================================================================
//======================                        ================================
//======================    POLY2 FUNCTIONS     ================================
//======================                        ================================
//==============================================================================
// 2D Polynomial Forms: describe any polynomial of 2 variables (a 'poly2') by
// its matrix of coefficients, given by  f(u,v) = [V]*[C]*[U], where 
//  V = [1, v, v^2, v^3, ...]  
//  U = [1, u, u^2, u^3, ...]transpose, and C = coefficient matrix. 
// In this matrix form it is easy to add, multiply, differentiate, or integrate
// polynomials. 

VECTYPE Matrx::poly2Eval(VECTYPE u0, VECTYPE v0)
//------------------------------------------------------------------------------
// Evaluate the poly2 described by 'this' at the point (u0,v0).
{
static Matrx uvec,vvec;
int i,j,imax,jmax;
VECTYPE val,tmp,out;

	imax = this->getXsize();			// make [1,u,u%2,...]^T matrix.
	uvec.sizer(imax,1);
	uvec.put0(0,0,1.0);
	val = 1.0;
	for(i=1; i<imax; i++)
	{
		val *= u0;
		uvec.put0(i,0,val);
	}
	jmax = this->getYsize();			// make [1 v v^2...] matrix.
	vvec.sizer(1,jmax);
	vvec.put0(0,0,1.0);
	val = 1.0;
	for(j=1; j<jmax; j++)
	{
		val *= v0;
		vvec.put0(0,j,val);
	}
	out = 0.0;
	for(j=0; j<jmax; j++)				// multiply (more efficient than mmpy)
	{
		val = 0.0;
		for(i=0; i<imax; i++)			// DOT product(uvec, j-th row of 'this')
		{
			val += get0(j,i)*uvec.get0(i,0);
		}
		tmp = vvec.get0(0,j);			// the v-vector element that weights it,
		out +=val*tmp;					// better precision:accum. row-by-row
	}
	return(out);
	
}

void Matrx::convolve(Mat3& inA3, Mat3& inB3)
//------------------------------------------------------------------------------
// inA, inB are coefficient matrices of two polynomials; find the 
// the product of these two polynomials by convolving their coef. matrices.
{
int i,j;
VECTYPE wt;

	sizer(5,5);						// Make sure we're the right size.
	zero();							// clear out current contents.
	for(i=0; i<3; i++)				// for each element of inB,
	{
		for(j=0; j<3; j++)			// (i=column, j=row)
		{
			wt = inB3.colrow[i][j];	// accumulate a shifted, scaled copy of inA
			addTo0(j  ,i  , wt * inA3.colrow[0][0]);
			addTo0(j+1,i  , wt * inA3.colrow[0][1]);
			addTo0(j+2,i  , wt * inA3.colrow[0][2]);
			addTo0(j  ,i+1, wt * inA3.colrow[1][0]);
			addTo0(j+1,i+1, wt * inA3.colrow[1][1]);
			addTo0(j+2,i+1, wt * inA3.colrow[1][2]);
			addTo0(j  ,i+2, wt * inA3.colrow[2][0]);
			addTo0(j+1,i+2, wt * inA3.colrow[2][1]);
			addTo0(j+2,i+2, wt * inA3.colrow[2][2]);
		}
	}
}

void Matrx::convolve(Matrx& inA, Matrx& inB) 
//------------------------------------------------------------------------------
// (faster: size of A LESS THAN size of B)
// Replace our current contents with result of convolving 2 matrices A, B; 
// if sizes of A, B are mxn and pxq, then size of result is (m+p),(n+q).
// Use convolution to find the product of 2-variable polynomials. 
// Example:
// define two polynomials of u,v by arranging their coefficients in a matrix;
// if K(u,v) = a + bu + cv + duv + eu^2v, (or any other polynomial of u & v),
//	K(u,v) = [1 v]*[a b 0]*[1  ]	Define the 2x3 matrix of coeffs as 'K'.
//                 [c d e] [u  ]
//                         [u^2]
// Similarly, arrange the coeffs of another polynomial R(u,v) to form matrix R.
// Here's the nifty part:  the coeffs of R(u,v)*K(u,v) are given by the matrix
// that results from 2D CONVOLUTION of matrices R and K, where the 'origin' of
// each matrix is defined as its upper-leftmost element.
// Uses a 'temp' matrix to store intermediate results, so a matrix object can
// use ITSELF as an argument to Matrx::convolve().
{
Matrx* pOut;
static Matrx tmp;					//(static avoids repeated malloc/free cycles)
int i,j,p,q,imax,jmax,pmax,qmax;
VECTYPE val;
BOOL useTmp;						// true if output held in temp. storage.

	imax = inA.getXsize();			// # columns in matrix A,
	jmax = inA.getYsize();			// # rows 
	pmax = inB.getXsize();			// # columns in matrix B,
	qmax = inB.getYsize();			// # rows.
	pOut = this;					// Point to ourselves as the output matrx;
	useTmp = FALSE;					// assume no temporary output needed.
	if(&inA==this || &inB==this)	// If output overwrites an input,
	{
		pOut = &tmp;				// use our local temporary output matrix,
		useTmp=TRUE;				// and set flag so we copy it to output.
	}
	pOut->sizer(jmax+qmax-1, imax+pmax-1);	// set output matrix size.
	pOut->zero();					// Set output matrix to zero to accumulate:
	for(i=0; i<imax; i++)
	{								
		for(j=0; j<jmax; j++)
		{							// for the j-th column, i-th row of A,
			val = inA.get0(j,i);
			for(p=0; p<pmax; p++)
			{
				for(q=0; q<qmax; q++)	// accum. [B]*val offset by j,i
				{
					pOut->addTo0(q+j,p+i,inB.get0(q,p)*val);
				}
			}
		}
	}
	if(useTmp==TRUE)				// if answer is in temporary storage,
	{
		copy(*pOut);				// replace 'this' with temp. contents,
	}
}

void Matrx::makeF3_hat(Mat3& F3, Mat2& Bx, Mat2& By)
//------------------------------------------------------------------------------
// Construct the \hat{F3} matrix (described in JGT 2003 "Quadrilateral.." paper 
// from F3, the coef matrix of the bi-quadratic 2D polynomial F3(x,y), and the
// bilinear mapping function described by the 2x2 matrices Bx,By, that convert
// an arbitrary quadrilateral (4-sided polygon) to the unit square.  The result
// is the 5x5 matrix \hat{F3} that holds the coefficients of f(x(u,v),y(u,v)). 
{
Mat3 Bx2,By2,Bxy;
Mat4 Bx2y,Bxy2;
static Matrx Bx2y2;					// (static avoids repeated new/delete ops )
int i,j;
VECTYPE wt;

	sizer(5,5);						// set final output matrix size, and
	zero();							// zero out all elements for accumulation.
	put0(0,0,F3.colrow[0][0]);		//-----f00 (constant coefficient of F3(x,y))
	wt = F3.colrow[1][0];			//-----f01 (coeff. for x term in F3(x,y))
	for(i=0; i<2; i++)		
	{
		for(j=0; j<2; j++)			// (i=row, j=column)
		{
			addTo0(i,j, Bx.colrow[j][i] * wt);	// add x(u,v) *f01
		}
	}
	wt = F3.colrow[2][0];			//-----f02 (coef for x^2 term in F3(x,y))
	Bx2.convolve(Bx,Bx);			// find x^2 in u,v terms
	for(i=0; i<3; i++)
	{
		for(j=0; j<3; j++)			// (i=row, j=column)
		{
			addTo0(i,j, Bx2.colrow[j][i] * wt);	// add x^2(u,v) *f02
		}
	}
	wt = F3.colrow[0][1];			//-----f10 (coeff. for y term in F3(x,y))
	for(i=0; i<2; i++)
	{
		for(j=0; j<2; j++)
		{
			addTo0(i,j, By.colrow[j][i] * wt);	// add y(u,v) *f10
		}
	}
	wt = F3.colrow[1][1];			//-----f11 (coeff. for xy term in F3(x,y))
	Bxy.convolve(Bx,By);			// find xy in u,v terms
	for(i=0; i<3; i++)
	{
		for(j=0; j<3; j++)
		{
			addTo0(i,j, Bxy.colrow[j][i] * wt);	// add x(u,v)y(u,v) *f11
		}
	}
	wt = F3.colrow[2][1];			//-----f12 (coeff. for x^2y term in F3(x,y))
	Bx2y.convolve(By,Bx2);			// find x^2y in u,v terms
	for(i=0; i<4; i++)
	{
		for(j=0; j<4; j++)
		{
			addTo0(i,j, Bx2y.colrow[j][i] * wt); // add x^2(u,v)y(u,v) *f12
		}
	}
	wt = F3.colrow[0][2];			//-----f20 (coeff. for y^2 term in F3(x,y))
	By2.convolve(By,By);			// find y^2 in u,v, terms
	for(i=0; i<3; i++)
	{
		for(j=0; j<3; j++)
		{
			addTo0(i,j, By2.colrow[j][i] * wt); // add y^2(u,v) *f20
		}
	}
	wt = F3.colrow[1][2];			//-----f21 (coeff. for xy^2 term in F3(x,y))
	Bxy2.convolve(Bx,By2);			// find xy^2 in u,v terms
	for(i=0; i<4; i++)
	{
		for(j=0; j<4; j++)
		{
			addTo0(i,j, Bxy2.colrow[j][i] * wt); // add x(u,v)y^2(u,v) *f21
		}
	}
	wt = F3.colrow[2][2];			//-----f22 (coef.for x^2y^2 term in F3(x,y))
	Bx2y2.convolve(Bx2,By2);
	for(i=0; i<5; i++)
	{
		for(j=0; j<5; j++)
		{
			addTo0(i,j, Bx2y2.get0(i,j) * wt); // add x^2(u,v)y^2(u,v) *f22
		}
	}
	// and we're done!
}

void Matrx::poly2Pow(int pow, Matrx& inA)
//------------------------------------------------------------------------------
// J. Tumblin 1/30/2002
// If 'inA' is the coefficient matrix of the 2D polynomial f(x,y), then find the 
// coefficient matrix that results from raising that polynomial to the positive
// integer power 'pow' like this: f(x,y)^pow.  How?  Convolve inA with itself
// (pow-1) times.  If inA matrix size is NxM, then result size is 
// (pow*(N-1)+1) X (pow*(M-1)+1).
// Each convolution puts its output at pOut, then copies pOut to pPrev to get
// ready for the next convolution.
// ****NOTE THAT THERE MIGHT BE FASTER WAYS TO DO THIS USING THE FFT*****
{
Matrx *pOut;
static Matrx tmp;					// static to avoid repeated malloc/free
static Matrx Mprev;	
int i,j,p,s,t,imax,jmax,smax,tmax;
VECTYPE val;
BOOL useTmp;						// true if output held in temp. storage.

#if(DIAGNOSTICS==TRUE)
	ASSERT(pow>1);					// !! must raise to power of 2 or more!!
#endif

	imax = inA.getXsize();			// input coefficient matrix size: # columns,
	jmax = inA.getYsize();			// # rows.
	pOut = this;					// Point to ourselves as the output matrx;
	useTmp = FALSE;					// Assume no temporary output needed, but
	if(&inA==this)					// if output overwrites the input,
	{
		pOut = &tmp;				// use a temporary output matrix,
		useTmp=TRUE;				// and set flag so we save it when done.
	}
	pOut->sizer(1+pow*(jmax-1),
				1+pow*(imax-1));	// set final output matrix size,
	pOut->zero();					// and clear the output matrix.
	for(i=0; i<imax; i++)			// Copy input into output (i.e. calc inA^1).
	{
		pOut->copy0Col(i,inA,i);	// (do not re-size pOut--just copy elements)
	}
	smax = tmax = 1;				// init to get proper size of Mprev matrix;
	for(p=1; p<pow; p++)			// For every power >1 of the polynomial,
	{
		Mprev.copy(*pOut);			// copy previous result, and then
		smax += imax-1;				// update the size of Mprev matrix (#cols,
		tmax += jmax-1;				// # of rows).
		for(i=0; i<imax; i++)		// convolve it with the input matrix;
		{								
			for(j=0; j<jmax; j++)
			{							// for the j,i-th element of A,
				val = inA.get0(j,i);
				for(s=0; s<smax; s++)
				{
					for(t=0; t<tmax; t++)	// accumulate a copy of [pPrev]
											// scaled by val & offset by j,i
					{
						pOut->addTo0(t+j,s+i,Mprev.get0(t,s)*val);
					}
				}
			}
		}
	}
	if(useTmp==TRUE)				// if answer is in temporary storage,
	{
		copy(*pOut);				// replace 'this' with temp. contents,
	}
}

VECTYPE Matrx::poly2quadIntegrateF3(Mat3& F3, 
									Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3)
//------------------------------------------------------------------------------
// Given any 2D bi-quadratic polynomial F3(x,y) (via its coef. matrix F3),
// compute and return the value of the definite integral of F3 within the
// quadrilateral region whose corner points in x,y are given by the counter-
// clockwise sequence of x,y points p0,p1,p2,p3, using the matrix-only method 
// described in the JGT 2004 submission "Quadrilateral Integration inside 
// Quadrilateral Boundaries". Note that these 4 points must form a CONVEX 
// quadrilateral--no chevrons or triangles!
{
static Mat2 Bx,By;					// for bilinear mapping from x,y to u,v 
static Matrx J;						// coefs of the Jacobian determinant
static Matrx Fhat;					// /hat{F} coeff. matrix.
VECTYPE sTmp,acc,out;
int i,j;

#define aCOEF	Bx.colrow[0][0]		// in paper, this is lowercase a coeff.
#define bCOEF	Bx.colrow[1][0]		// b
#define cCOEF	Bx.colrow[0][1]		// c
#define dCOEF	Bx.colrow[1][1]		// d.
#define eCOEF	By.colrow[0][0]		// e
#define fCOEF	By.colrow[1][0]		// f
#define gCOEF	By.colrow[0][1]		// g
#define hCOEF	By.colrow[1][1]		// h.

		// Build the Bx, By matrices from corner pts
	aCOEF =  p0.row[0];										// a
	bCOEF = -p0.row[0] +p1.row[0];							// b
	cCOEF = -p0.row[0]			             +p3.row[0];	// c
	dCOEF =  p0.row[0] -p1.row[0] +p2.row[0] -p3.row[0];	// d. ............
	eCOEF =  p0.row[1];										// e
	fCOEF = -p0.row[1] +p1.row[1];							// f
	gCOEF = -p0.row[1]						 +p3.row[1];	// g
	hCOEF =  p0.row[1] -p1.row[1] +p2.row[1] -p3.row[1];	// h. .............
//Bx.prnt("Bx matrix");
//By.prnt("By matrix");
									// Build the J matrix from the corner points
	J.sizer(2,2);				// (be sure it's the right size)
	J.put0(0,0, (gCOEF * bCOEF) - (fCOEF * cCOEF));	// row 0, col 0
	J.put0(0,1, (hCOEF * bCOEF) - (fCOEF * dCOEF));	// row 0, col 1
	J.put0(1,0, (gCOEF * dCOEF) - (hCOEF * cCOEF));	// row 1, col 0
	J.put0(1,1, (VECTYPE)0.0);						// row 1, col 1.
//J.prnt("J matrix");
	Fhat.makeF3_hat(F3,Bx,By);	// Build the F_hat matrix (slow!)
//Fhat.prnt("Fhat matrix");
	convolve(J,Fhat);			// do the convolution with J. Result is 6x6
//prnt("J*Fhat matrix");
	out = (VECTYPE)0.0;
	for(j=0; j<6; j++)			// Do the 'bigOdot' operator with SS and sum;
	{							// (for each row,...)
								// (for computing each element of SS^T: )
		sTmp = (VECTYPE)1.0 / ((VECTYPE)(j+1));	
		acc = (VECTYPE)0.0;		// (row accumulator)
		for(i=0; i<6; i++)		// (for each column...)
		{
			acc += get0(j,i) * sTmp/((VECTYPE)(i+1));
		}
		out += acc;				// accumulate row results
		// (NOTE: why use 'acc' at all--why not just accumulate in 'out'?
		//  As 'out' gets large, accumulating small values from each element
		//  has less and less precision. 'acc' values are usually very close
		//  to individual values (exponents very similar); adds more precise).
	}
	return(out);				// and that's the answer!
	
#undef aCOEF						// Locals only!
#undef bCOEF
#undef cCOEF
#undef dCOEF
#undef eCOEF
#undef fCOEF
#undef gCOEF
#undef hCOEF
}

VECTYPE Matrx::poly2quadMC(Matrx& F3, Vec2& p0, Vec2& p1, Vec2& p2, Vec2& p3)
//------------------------------------------------------------------------------
// Evaluate the integral within the quadrilateral defined by 4 points by using
// jittered supersampling.
{
#define MAXcount 4000000		// desired max# of samples to shoot.
int i,j,imax,jmax;				// grid cell indexing.
VECTYPE xmin,xmax,ymin,ymax, xrng,yrng;
VECTYPE acc,out;
Vec2 sampl;
double xcel,ycel;				// size of grid cel that holds a 
double ybase;					// randomly-placed sample.
int cls;						// quadrilateral classification.

								// find a bounding box for the quadrilateral:
	xmin = xmax = p0.row[0]; 
	ymin = ymax = p0.row[1];
	if(p1.row[0] < xmin) xmin = p1.row[0];
	if(p2.row[0] < xmin) xmin = p2.row[0];
	if(p3.row[0] < xmin) xmin = p3.row[0];

	if(p1.row[0] > xmax) xmax = p1.row[0];
	if(p2.row[0] > xmax) xmax = p2.row[0];
	if(p3.row[0] > xmax) xmax = p3.row[0];

	if(p1.row[1] < ymin) ymin = p1.row[1];
	if(p2.row[1] < ymin) ymin = p2.row[1];
	if(p3.row[1] < ymin) ymin = p3.row[1];

	if(p1.row[1] > ymax) ymax = p1.row[1];
	if(p2.row[1] > ymax) ymax = p2.row[1];
	if(p3.row[1] > ymax) ymax = p3.row[1];
	xrng = xmax - xmin;
	yrng = ymax - ymin;
	// For spatially uniformity, use jittered supersampling.  Divide xrng,yrng
	// into MAXcount square grid cells, then make one sample at a random
	// position within each gridcell.
	// to find imax,jmax, use:  xrng/yrng = imax/jmax; imax*jmax = MAXcount.
	//   (jmax=MAXcount/imax, so (MAXcount/imax)(xrng/yrng) = imax, so...
	imax = (int)sqrt(MAXcount*((double)xrng/(double)yrng));
	jmax = (int)((double)MAXcount/(double)imax);
	xcel = xrng / (double)imax;
	ycel = yrng / (double)jmax;
	out = (VECTYPE)0.0;			
	for(j=0; j<jmax; j++)		// (rows of cells)
	{
		if(j%50==0) dprintf(".");			// timekeeper
		acc = (VECTYPE)0.0;		// sub-accumulator for better accuracy
		ybase = ymin + yrng*((VECTYPE)j/(VECTYPE)(jmax-1)); 
								// bottom of current row of cells.
		for(i=0; i<imax; i++)	// (cells within a row)
		{						// choose random spot within the bounding box
			sampl.row[0] = xmin  + xrng*((VECTYPE)i/(VECTYPE)(imax-1))
								 + xcel*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
			sampl.row[1] = ybase + ycel*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
			if(TRUE==sampl.isInsideQuad(p0,p1,p2,p3))
			{					// if it's inside the quadrilateral, then
				acc += F3.poly2Eval(sampl.row[0],sampl.row[1]);	
			}					// add it to our sub-accumulator;
								// if not, it's zero-valued
		}
		out += (acc / (VECTYPE)imax)*xrng;	// accumulate this row's results.
	}
	out = (out / (VECTYPE)jmax)*yrng;		// accumulate row results.
	dprintf("\n");				// next line.
	//!!! SUBTLETY!!!----------
	// The method used here IGNORES the order of the vertices; e.g. there
	// is no notion 'backfacing/frontfacing' quads, as there is for the matrix
	// based solution. Use Vec2::classifyQuads() to set the sign of result:
	cls = sampl.classifyQuad(p0,p1,p2,p3);
	if(cls<0) out = -out;
	return(out);

#undef MAXcount					// Locals Only!
}

void Matrx::poly2convexQuadTestPrnt(void)
//------------------------------------------------------------------------------
// Test poly2quadIntegrateF3() function by comparing its answers against
//	    poly2quadMC(); generate random CONVEX quadrilaterals and random 3x3
// matrices for F3, compare the results and print out any that cause large
// errors.
{
#define MAXtest	100000			// number of random polynomials to test
int i;
Mat3 F3test;
Matrx F3m;						// re-sizeable version of F3test.
Vec2 p0,p1,p2,p3;				// quadrilateral corners.
Vec2 axis,v1,v2;
Mat2 rot;
VECTYPE err2,ansQuad, ansMC;
int isBad;						// Quad classification.

	F3m.sizer(3,3);
	srand(time(NULL));			// seed the random number generator, and...
	// TEST: Unit Square for the quadrilateral:
//	p0.row[0] = (VECTYPE)0.0; p0.row[1] = (VECTYPE)0.0;
//	p1.row[0] = (VECTYPE)1.0; p1.row[1] = (VECTYPE)0.0;
//	p2.row[0] = (VECTYPE)1.0; p2.row[1] = (VECTYPE)1.0;
//	p3.row[0] = (VECTYPE)0.0; p3.row[1] = (VECTYPE)1.0;
	for(i=1; i<MAXtest; i++)
	{
		// make random quadrilateral; x,y coords within -10 to +10:
			// make 2 opposite corners;
		p0.row[0] = -10.0 + 20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX); 
		p0.row[1] = -10.0 +	20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
		p2.row[0] = -10.0 + 20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX); 
		p2.row[1] = -10.0 +	20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
			// then put p1 on the right side of the p0,p2 line;
		axis.sub(p2,p0);
		axis.norm(axis);			// Normalize its length;
		rot.makeRot270();		
		v1.mpy(rot,axis);		// Find random place using v1,axis coordinates
		v1.scale(v1, 3.0*(VECTYPE)rand()/(VECTYPE)RAND_MAX); // 
		axis.scale(axis, -3.0 + 3.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX));
		p1.add(axis,v1);
		p1.add(p1,p0);			// (position is relative to p0).
			// Tricky process chooses p3 so that quadrilateral is convex:
			// let r be random var between 0 and 1.  Then
			// p3 = p0 + (p2-p0)r + r*scale*(p2-p1) + r*scale*(p0-p1)
			//    = p0 + axis*r + r*scale*v1 + r*scale*v2
		axis.sub(p2,p0);
		axis.scale(axis,((VECTYPE)rand()/ RAND_MAX));
		v1.sub(p2,p1);
		v1.scale(v1,3.0*((VECTYPE)rand()/ RAND_MAX));
		v2.sub(p0,p1);
		v2.scale(v2,3.0*((VECTYPE)rand()/ RAND_MAX));
		v1.add(v1,v2);
		axis.add(axis,p0);
		p3.add(v1,axis);			// (position is relative to p0).

//F3m.put0Row(0, 1.0,0.0,0.0);
//F3m.put0Row(1, 0.0,0.0,0.0);
//F3m.put0Row(2, 0.0,0.0,0.0);
									// Make a random 2-D cubic polynomial, but
		F3m.randFill(-10.0,10.0);	// keep coeffs between -10 and +10.
		F3m.copyTo(F3test);			// Make a copy of those coeffs.
//F3m.prnt("F3 Matrix:");		// print the matrix we made,
		ansMC = poly2quadMC(F3m,p0,p1,p2,p3);	// compute by method 1,
		ansQuad = poly2quadIntegrateF3(F3test,p0,p1,p2,p3);	// and by method 2.
		err2 = 2.0*(ansQuad-ansMC)/(ansQuad+ansMC);
		if(err2*err2 > 0.01*0.01)
		{
			isBad = axis.classifyQuad(p0,p1,p2,p3); // what's the quad's shape?
			F3m.prnt("ERROR > 1% on this matrix!");
			// Print out the coords of the quad:
			dprintf("p0=(%f, %f), p1=(%f, %f)\n",p0.row[0],p0.row[1],p1.row[0],p1.row[1]);
			dprintf("p2=(%f, %f), p3=(%f, %f)\n",p2.row[0],p2.row[1],p3.row[0],p3.row[1]);
			dprintf("And quad classification is %d \n",isBad);
		}
		// Always report the amount of error we found.
		dprintf("Convex quad--US: %f vs. MC: %f; err=%f\n",ansQuad,ansMC,err2);

	}
}
void Matrx::poly2arbitraryQuadTestPrnt(void)
//------------------------------------------------------------------------------
// Test poly2quadIntegrateF3() function by comparing its answers against
//	    poly2quadMC(); generate random ARBITRARY quadrilaterals and random 3x3
// matrices for F3, compare the results and print out any that cause large
// errors.
{
#define MAXtest	100000			// number of random polynomials to test
int i, isBad;
Mat3 F3test;
Matrx F3m;						// re-sizeable version of F3test.
Vec2 p0,p1,p2,p3;				// quadrilateral corners.
Vec2 axis,v1,v2;
Mat2 rot;
VECTYPE err2,ansQuad, ansMC;

	F3m.sizer(3,3);
	srand(time(NULL));			// seed the random number generator, and...
	// TEST: Unit Square for the quadrilateral:
//	p0.row[0] = (VECTYPE)0.0; p0.row[1] = (VECTYPE)0.0;
//	p1.row[0] = (VECTYPE)1.0; p1.row[1] = (VECTYPE)0.0;
//	p2.row[0] = (VECTYPE)1.0; p2.row[1] = (VECTYPE)1.0;
//	p3.row[0] = (VECTYPE)0.0; p3.row[1] = (VECTYPE)1.0;
	for(i=1; i<MAXtest; i++)
	{
		// make random quadrilateral; x,y coords within -10 to +10:
		p0.row[0] = -10.0 + 20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX); 
		p0.row[1] = -10.0 +	20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
		p1.row[0] = -10.0 + 20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX); 
		p1.row[1] = -10.0 +	20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
		p2.row[0] = -10.0 + 20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX); 
		p2.row[1] = -10.0 +	20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX);
		p3.row[0] = -10.0 + 20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX); 
		p3.row[1] = -10.0 +	20.0*((VECTYPE)rand()/(VECTYPE)RAND_MAX);

		// what is the shape of this quad?
		isBad = axis.classifyQuad(p0,p1,p2,p3);
		// Skip selected pathological shapes:
		// (comment out the shapes you will allow)
		if( 	isBad == JTQUAD_DEGEN		||isBad == JTQUAD_BOWTIE		
//			||	isBad == JTQUAD_CHEVRON_02	||isBad == JTQUAD_CHEVRON_13	
//			||	isBad == JTQUAD_TRI_012		||isBad == JTQUAD_TRI_123
//			||	isBad == JTQUAD_TRI_230		||isBad == JTQUAD_TRI_301
			)
			continue;

// TEST: easiest polynomial F(x,y) = 1:
//F3m.put0Row(0, 1.0,0.0,0.0);
//F3m.put0Row(1, 0.0,0.0,0.0);
//F3m.put0Row(2, 0.0,0.0,0.0);
									// Make a random 2-D cubic polynomial, but
		F3m.randFill(-10.0,10.0);	// keep coeffs between -10 and +10.
		F3m.copyTo(F3test);			// Make a copy of those coeffs.
//F3m.prnt("F3 Matrix:");		// print the matrix we made,
		ansMC = poly2quadMC(F3m,p0,p1,p2,p3);	// compute by method 1,
		ansQuad = poly2quadIntegrateF3(F3test,p0,p1,p2,p3);	// and by method 2.
		err2 = 2.0*(ansQuad-ansMC)/(ansQuad+ansMC);
		if(err2*err2 > 0.01*0.01)
		{
			F3m.prnt("ERROR > 1% on this matrix!");
			// Print out the coords of the quad:
			dprintf("p0=(%f, %f), p1=(%f, %f\n)",p0.row[0],p0.row[1],p1.row[0],p1.row[1]);
			dprintf("p2=(%f, %f), p3=(%f, %f\n)",p2.row[0],p2.row[1],p3.row[0],p3.row[1]);
			dprintf("And quad classification is %d \n",isBad);
		}
		// report the amount of error we found.
		dprintf("Arbitrary quad, type=%d; US: %f vs. MC: %f; err=%f\n",isBad, ansQuad,ansMC,err2);

	}
}

#undef PESKYWARNINGS
#undef DIAGNOSTICS

